/*  
 * Seward's algorithm d_copyEQ_u12.c modified in order to compute the
 * standard sa (not the cyclic one) and to allow 0's in the input
 * 
 * ver 1.0 (12-apr-00)  
 * ver 1.1 (26-mag-00) works with 0's in the input text
 */
#include <assert.h>
#include <stdio.h>

typedef int          Int32;
typedef unsigned int UInt32;
typedef char          Char;
typedef unsigned char UChar;
typedef unsigned char Bool;
#define True   ((Bool)1)
#define False  ((Bool)0)
#define N_OVERSHOOT (16-1)
#define BIGFREQ(b) (ftab[((b)+1) << 8] - ftab[(b) << 8])
#define SETMASK (1 << 30)
#define CLEARMASK (~(SETMASK))


extern int Suftest_verbose;
extern int Suftest_reverse_order;


Int32  nBlock;             // size of input string 
UChar  *block;             // input string+ overshoot
Int32  *ptr;               // suffix array
Int32  ftab [65537];      
Int32  runningOrder[256];


void suffixsort(UChar *x, int *p, int n)
{
  void d_copyEQ_u12( void );
  block=x;
  ptr=p;
  nBlock=n;
  d_copyEQ_u12();  
}


static 
__inline__
Int32 d_copyEQ_u14_cmp0( Int32 b1, Int32 b2 )
{
   UChar c1, c2;
   if (b1 == b2) return 0;

#if 0
   // 1
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 

   // 2
   c1 = block[b1]; c2 = block[b2];
   if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
   b1++; b2++; 
#else
   b1 += 2; b2 += 2;  // the first two chars are always equal
#endif

   // execute blocks of 14 comparisons untill a difference
   // is found or we run out of the string 
   do {
     // 3
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 4
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 5
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 6
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 7
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 8
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 
   
     // 9
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 10
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 11
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 12
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 13
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 14
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 15
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // 16
     c1 = block[b1]; c2 = block[b2];
     if (c1 != c2) return ((UInt32)c1 - (UInt32)c2);
     b1++; b2++; 

     // if (b1 >= nBlock) b1 -= nBlock;
     // if (b2 >= nBlock) b2 -= nBlock;
   } while( b1<nBlock && b2<nBlock);

   return b2-b1;   // we have  b2>b1 <=> s2<s1
} 


#if !MK
#define  TEMPLATE  d_copyEQ_u12
#define  CMP       d_copyEQ_u14_cmp0
#define  FMAP      ptr
#include "qsort3.template"
#endif


static
void calc_running_order ( void )
{
   Int32 i, j;
   for (i = 0; i <= 255; i++) runningOrder[i] = i;

   {
      Int32 vv;
      Int32 h = 1;
      do h = 3 * h + 1; while (h <= 256);
      do {
         h = h / 3;
         for (i = h; i <= 255; i++) {
            vv = runningOrder[i];
            j = i;
            while ( BIGFREQ(runningOrder[j-h]) > BIGFREQ(vv) ) {
               runningOrder[j] = runningOrder[j-h];
               j = j - h;
               if (j <= (h - 1)) goto zero;
            }
            zero:
            runningOrder[j] = vv;
         }
      } while (h != 1);
   }
   if(Suftest_reverse_order) {
     Int32 t;
     for(i=0;i<128;i++) {
       t=runningOrder[255-i];
       runningOrder[255-i]=runningOrder[i];
       runningOrder[i]=t;
     }
   }
}



void d_copyEQ_u12 ( void )
{
   void ssort1main(UChar *, int *, int);
   Int32  i, j, k, ss, sb;
   UChar  c1, c2;
   Bool   bigDone[256];
   Int32  copyStart[256];
   Int32  copyEnd  [256];
   Int32  numQSorted = 0;

   for (i = 0; i < N_OVERSHOOT; i++) // init tail
      block[nBlock + i] = 0;

   for (i = 0; i <= 65536; i++) ftab[i] = 0;

   c1 = block[0];
   for (i = 1; i <= nBlock; i++) {
      c2 = block[i];
      ftab[(c1 << 8) + c2]++;
      c1 = c2;
   } 

   for (i = 1; i <= 65536; i++) ftab[i] += ftab[i-1];

   c1 = block[0];
   for (i = 0; i < nBlock; i++) {
      c2 = block[i+1];
      j = (c1 << 8) + c2;
      c1 = c2;
      ftab[j]--;
      ptr[ftab[j]] = i;
   }

   /* decide on the running order */
   calc_running_order();
   for (i = 0; i < 256; i++) bigDone[i] = False;

   /* Really do the sorting */
   for (i = 0; i <= 255; i++) {
  
      /*--
         Process big buckets, starting with the least full.
      --*/
      ss = runningOrder[i];

      /*--
         Complete the big bucket [ss] by quicksorting
         any unsorted small buckets [ss, j].  Hopefully
         previous pointer-scanning phases have already
         completed many of the small buckets [ss, j], so
         we don't have to sort them at all.
      --*/
      for (j = 0; j <= 255; j++) {
         if (j != ss) {
            sb = (ss << 8) + j;
            if ( ! (ftab[sb] & SETMASK) ) {
               Int32 lo = ftab[sb]   & CLEARMASK;
               Int32 hi = (ftab[sb+1] & CLEARMASK) - 1;
               if (hi > lo) {
                  if (Suftest_verbose>1)
                     fprintf ( stderr,
                               "        qsort [0x%x, 0x%x]   done %d"
                               "   this %d\n",
                               ss, j, numQSorted, hi - lo + 1 );
                  #if MK
                  ssort1main(block, ptr+lo, hi-lo+1);
                  #else
                  d_copyEQ_u12_qsort3 ( lo, hi );
                  #endif
                  numQSorted += ( hi - lo + 1 );
               }
            }
            ftab[sb] |= SETMASK;
         }
      }

      assert (!bigDone[ss]);

      {
         for (j = 0; j <= 255; j++) {
            copyStart[j] =  ftab[(j << 8) + ss]     & CLEARMASK;
            copyEnd  [j] = (ftab[(j << 8) + ss + 1] & CLEARMASK) - 1;
         }
	 // take care of the virtual -1 char in position nBlock+1
	 if(ss==0) {
	   k=nBlock-1;
	   c1 = block[k];
	   if (!bigDone[c1])
	     ptr[ copyStart[c1]++ ] = k;
	 }
         for (j = ftab[ss << 8] & CLEARMASK; j < copyStart[ss]; j++) {
            k = ptr[j]-1; if (k < 0) continue;   // k += nBlock;
            c1 = block[k];
            if (!bigDone[c1])
               ptr[ copyStart[c1]++ ] = k;
         }
         for (j = (ftab[(ss+1) << 8] & CLEARMASK) - 1; j > copyEnd[ss]; j--) {
            k = ptr[j]-1; if (k < 0) continue; // k += nBlock;
            c1 = block[k];
            if (!bigDone[c1]) 
               ptr[ copyEnd[c1]-- ] = k;
         }
      }

      assert (copyStart[ss] - 1 == copyEnd[ss]);

      for (j = 0; j <= 255; j++) ftab[(j << 8) + ss] |= SETMASK;
      bigDone[ss] = True;

   }

   if (Suftest_verbose>0)
   fprintf ( stderr, "        %d pointers, %d sorted, %d scanned\n",
                     nBlock, numQSorted, nBlock - numQSorted );

}
